////////////////////////////////////////////////////////////////////////////////
//
// JPEGLoader.pas - JPEG wrapper
// -----------------------------
// Version:   2003-02-26
// Maintain:  Michael Vinther    |    mv@logicnet.dk
//
// Last changes:
//
unit JPEGLoader;

interface

uses SysUtils, LinarBitmap, Streams, Windows, Graphics, DelphiStream, JPEG, MemStream,
  Classes, Monitor;

resourcestring
  rsJPEGImageFile = 'Joint Picture Expert Group';

type
  TJPEGLoader = class(TBitmapLoader)
    public
      Quality       : Integer; // Image quality, 1=worst, 100=best
      FileSizeLimit : Integer; // If FileSizeLimit>0, SaveToStream will try to limit the file size
                               // to FileSizeLimit bytes.

      function CanLoad(const Ext: string): Boolean; override;
      function CanSave(const Ext: string): Boolean; override;

      function GetLoadFilter: string; override;
      function GetSaveFilter: string; override;

      procedure LoadFromStream(Stream: TSeekableStream; const Ext: string; Bitmap: TLinarBitmap); override;
      procedure SaveToStream(Stream: TSeekableStream; const Ext: string; LBitmap: TLinarBitmap); override;
    end;

var
  Default : TJPEGLoader;

// Convert to grayscale image
procedure ConvertToGrayscale(const Image,GrayImage: TLinarBitmap); overload;
procedure ConvertToGrayscale(const Image: TLinarBitmap); overload;

implementation

// Convert to grayscale image
procedure ConvertToGrayscale(const Image,GrayImage: TLinarBitmap);
var
  Pix : ^RGBRec;
  NewPix : ^Byte;
  P : Integer;
begin
  if Image.PixelFormat=pf24bit then
  begin
    GrayImage.New(Image.Width,Image.Height,pf8bit);
    GrayImage.Palette^:=GrayPal;
    Pointer(Pix):=Image.Map;
    Pointer(NewPix):=GrayImage.Map;
    for P:=0 to GrayImage.Size-1 do
    begin
     NewPix^:=(Pix^.R*2+Pix^.G*3+Pix^.B+3) div 6;
     Inc(Pix); Inc(NewPix);
    end;
  end
  else
  begin
    Image.AssignTo(GrayImage);
    ConvertToGrayscale(GrayImage);
  end;
end;

// Convert to grayscale image
procedure ConvertToGrayscale(const Image: TLinarBitmap);
var
  Pix : ^Byte;
  P : Integer;
  LUT : array[0..255] of Byte;
  NewImage : TLinarBitmap;
begin
  if Image.PixelFormat=pf24bit then
  begin
    NewImage:=TLinarBitmap.Create;
    ConvertToGrayscale(Image,NewImage);
    Image.TakeOver(NewImage);
    NewImage.Free;
  end
  else
  begin
    for P:=0 to 255 do LUT[P]:=(Image.Palette^[P].R*2+Image.Palette^[P].G*3+Image.Palette^[P].B+3) div 6;
    Pointer(Pix):=Image.Map;
    for P:=1 to Image.Size do
    begin
      Pix^:=LUT[Pix^];
      Inc(Pix);
    end;
    Image.Palette^:=GrayPal;
  end;
end;

function TJPEGLoader.GetLoadFilter: string;
begin
  Result:=rsJPEGImageFile+' (*.jpg;*.jpeg;*.jpe)|*.jpg;*.jpeg;*.jpe';
end;

function TJPEGLoader.GetSaveFilter: string;
begin
  Result:=rsJPEGImageFile+' (*.jpg)|*.jpg';
end;

function TJPEGLoader.CanLoad(const Ext: string): Boolean;
begin
  Result:=(Ext='JPG') or (Ext='JPEG') or (Ext='JPE');
end;

function TJPEGLoader.CanSave(const Ext: string): Boolean;
begin
  Result:=(Ext='JPG') or (Ext='JPEG');
end;

procedure TJPEGLoader.LoadFromStream(Stream: TSeekableStream; const Ext: string; Bitmap: TLinarBitmap);
var
  JPEG : TJPEGImage;
  DelphiStream : TSeekableDelphiStream;
begin
  JPEG:=TJPEGImage.Create;
  try
    DelphiStream:=TSeekableDelphiStream.Create(Stream);
    try
      JPEG.LoadFromStream(DelphiStream);
    finally
      DelphiStream.Free;
    end;
    JPEG.Performance:=jpBestQuality;
    JPEG.Scale:=jsFullSize;
    JPEG.PixelFormat:=jf24bit;
    Bitmap.Assign(JPEG);
    if JPEG.GrayScale then ConvertToGrayscale(Bitmap);
  finally
    JPEG.Free;
  end;
end;

procedure TJPEGLoader.SaveToStream(Stream: TSeekableStream; const Ext: string; LBitmap: TLinarBitmap);
var
  X, Y : Integer;
  Bitmap : TBitmap;
  JPEG : TJPEGImage;
  OutPix : ^RGBRec;
  Pix : ^Byte;
  DelphiStream : TSeekableDelphiStream;
  MemStream : TMemoryStream;
  MinAboveSizeQuality, MaxBelowSizeQuality, CurrentSize, QualityGuess : Integer;
begin
  Bitmap:=TBitmap.Create;
  try
    if LBitmap.Pixelformat=pf24bit then LBitmap.AssignTo(Bitmap)
    else if LBitmap.Pixelformat=pf8bit then
    begin
      Bitmap.PixelFormat:=pf24bit;
      Bitmap.Width:=LBitmap.Width;
      Bitmap.Height:=LBitmap.Height;
      for Y:=0 to LBitmap.Height-1 do
      begin
       Pix:=@LBitmap.ScanLine[Y]^;
       OutPix:=@Bitmap.ScanLine[Y]^;
       for X:=0 to LBitmap.Width-1 do
       begin
        OutPix^:=LBitmap.Palette^[Pix^];
        Inc(Pix); Inc(OutPix);
       end;
      end;
    end;

    JPEG:=TJPEGImage.Create;
    try
      JPEG.Scale:=jsFullSize;
      JPEG.PixelFormat:=jf24bit;
      if LBitmap.IsGrayscale then JPEG.GrayScale:=True;

      if FileSizeLimit>0 then // Optimize file size to FileSizeLimit
      begin
        if Quality<1 then Quality:=JPEG.CompressionQuality;
        QualityGuess:=Quality;
        MinAboveSizeQuality:=Quality+1;
        MaxBelowSizeQuality:=1;
        MemStream:=TMemoryStream.Create;
        try
          repeat
            MemStream.Position:=0;
            JPEG.CompressionQuality:=QualityGuess;
            JPEG.Assign(Bitmap);
            JPEG.SaveToStream(MemStream); 
            CurrentSize:=MemStream.Position;
            if CurrentSize>FileSizeLimit then
            begin
              MinAboveSizeQuality:=QualityGuess; 
              if QualityGuess<=1 then Break;
            end
            else if CurrentSize<FileSizeLimit then
            begin
              MaxBelowSizeQuality:=QualityGuess;
              if (QualityGuess>=100) or (MaxBelowSizeQuality+1=MinAboveSizeQuality) then Break;
            end
            else Break; // CurrentSize=FileSizeLimit
            QualityGuess:=(MinAboveSizeQuality+MaxBelowSizeQuality) div 2;
          until False;

          Stream.Write(MemStream.Memory^,CurrentSize);
        finally
          MemStream.Free;
        end;
      end
      else // Use specified quality
      begin
        if Quality>=1 then JPEG.CompressionQuality:=Quality;
        JPEG.Assign(Bitmap);
        DelphiStream:=TSeekableDelphiStream.Create(Stream);
        try
          JPEG.SaveToStream(DelphiStream);
        finally
          DelphiStream.Free;
        end;
      end;
    finally
      JPEG.Free;
    end;
  finally
    Bitmap.Free;
  end;
end;

initialization
  Default:=TJPEGLoader.Create;
  LinarBitmap.AddLoader(Default);
finalization
  Default.Free;
end.

